/*************************************************************************
*                                                                        *
* 	XMMHandler - Access XMM for storing / accessing db_Vista records		 *
* 																								 *
* 	This file implements the routines that the user calls. This library	 *
* 	calls the modified XMM routines (for large model) as distributed by	 *
* 	Microsoft.                                                         	 *
* 																								 *
* 	Author:		Tom Frank, BIMCO														 *
*                                                                        *
*************************************************************************/

#ifndef	LINT_ARGS
#define	LINT_ARGS
#endif

#ifndef	ANSI
#define	ANSI
#endif

#include	"xmm.h"
#include "xmmhndlr.h"
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>

/*************************************************************************
*                                                                        *
* 	Defines used to return error codes - these match the db_Vista			 *
* 	conventions - here so we don't need vista.h									 *
*                                                                        *
*************************************************************************/

#define	S_OKAY	0
#define	S_NOTFOUND	2

/*************************************************************************
*                                                                        *
* 	Defines used within this module													 *
*                                                                        *
*************************************************************************/

#define	TRUE	1
#define	FALSE	0
#define	UNUSED	-1
#define	ONE_K		1024L
#define	MAX_REC_LEN	256

/*************************************************************************
*                                                                        *
* 	Data Structure Definitions that are not visible outside this module	 *
*                                                                        *
*************************************************************************/

typedef struct _XMMHandle {						/* XMM handle data structure */
	short handle;										/* XMM handle */
	int adj_rec_size;									/* user's rounded record size */
	int act_rec_size;									/* user's actual record size */
	long max_rec_alloc;								/* highest rec # alloc */
	long max_rec_used;								/* highest rec # used */
	} XMMHandle, *XMMHandlePTR;

/*************************************************************************
*                                                                        *
* 	Global Data Available from this module											 *
*                                                                        *
*************************************************************************/

int XMMErrorCode;										/* return code from XMM rtns */

/*************************************************************************
*                                                                        *
* 	Local Data Storage - internal to this module									 *
*                                                                        *
*************************************************************************/

static XMMHandlePTR XMM_Handle_Table;			/* pointer to handle table */
static int initialized = FALSE;					/* started?? */
static int Max_Handle_Alloc;						/* User requested max handles */
static int Max_Handle_Used;						/* last handle used */
static char move_area[MAX_REC_LEN];				/* for data movement */

/*************************************************************************
*                                                                        *
* 	Local Functions internal to this module										 *
*                                                                        *
*************************************************************************/

static int null_record(char *buffer, int buff_len);

/*<f>*/
int InitXMMSystem(int MaxHandles)
/*************************************************************************
*                                                                        *
* 	Initialize the XMM interface system - return TRUE / FALSE to signify	 *
* 	our success level																		 *
*                                                                        *
*************************************************************************/
{
	int i;

	if (initialized)									/* already done?? */
		return (S_NOTFOUND);							/* yes */

	if (!XMM_Installed())							/* is XMM handler around?? */
		return (S_NOTFOUND);							/* no */

	if ((XMM_Handle_Table=(XMMHandlePTR)malloc(sizeof(XMMHandle) * MaxHandles)) == NULL)
		return (S_NOTFOUND);							/* no memory */

	memset((char *)XMM_Handle_Table, '\0', sizeof(XMMHandle) * MaxHandles);

	for (i=0; i < MaxHandles; i++)				/* set all as unused */
		XMM_Handle_Table[i].handle = UNUSED;

	initialized = TRUE;								/* started now */
	Max_Handle_Alloc = MaxHandles;				/* User requested max handles */
	Max_Handle_Used = -1;							/* last handle used (none yet) */

	return (S_OKAY);

}

/*<f>*/
int ShutXMMSystem(void)
/*************************************************************************
*                                                                        *
* 	Close down the XMM interface system - return S_OKAY / S_NOTFOUND to 	 *
* 	signify our success level															 *
*                                                                        *
*************************************************************************/
{
	int i;

	if (!initialized)									/* already done?? */
		return (S_NOTFOUND);							/* yes */

/*************************************************************************
*                                                                        *
* 	Insure all Handles have been released											 *
*                                                                        *
*************************************************************************/

	for (i=0; i <= Max_Handle_Used; i++)
		{
		if (XMM_Handle_Table[i].handle != UNUSED)	/* in use?? */
			XMM_FreeExtended(XMM_Handle_Table[i].handle);
		}

/*************************************************************************
*                                                                        *
* 	Drop memory for Handle Table														 *
*                                                                        *
*************************************************************************/

	free(XMM_Handle_Table);

/*************************************************************************
*                                                                        *
* 	Reset Flags																				 *
*                                                                        *
*************************************************************************/

	initialized = FALSE;								/* not started now */

	return (S_OKAY);

}

/*<f>*/
XHandle AllocateXMMArea(long MaxRecs, int RecSize)
/*************************************************************************
*                                                                        *
* 	Allocate an XMM area and return the handle to the user. A negative	 *
* 	return is an error and the actual error (if -3) is in XMMErrorCode	 *
*                                                                        *
* 	Note: we allocate 1 more record than MaxRecs since 0 is a valid recno *
*                                                                        *
*************************************************************************/
{
	unsigned space_needed;
	long ret_code;
	XHandle hnd;
	struct	XMM_Move CtrlBlk;
	char block[ONE_K];
	int i;
	int adj_recsize;

	XMMErrorCode = 0;									/* clear code */

	if (!initialized)									/* started?? */
		return (-1);									/* no */

	for (hnd=-1,i=0; i < Max_Handle_Alloc; i++)
		if (XMM_Handle_Table[i].handle == UNUSED)/* available?? */
			{
			hnd = i;										/* yes */
			break;
			}

	if (hnd == -1)										/* space left in table?? */
		return (-2);									/* no */

	if (RecSize > MAX_REC_LEN)						/* asking for too big a record?? */
		return (-5);									/* yes */

	adj_recsize = RecSize;							/* copy this */
	if (adj_recsize & 0x0001)						/* is it odd? */
		adj_recsize++;									/* make even */

	ret_code =((long)(MaxRecs + 1L) * (long)adj_recsize + (ONE_K - 1L)) / ONE_K;
	space_needed = (unsigned)ret_code;
	ret_code = XMM_AllocateExtended(space_needed);
	if (ret_code < 0L)								/* error?? */
		{
		XMMErrorCode = XMSERROR(ret_code);		/* yes - save error */
		return (-3);									/* and leave */
		}

	XMM_Handle_Table[hnd].handle = (short)ret_code;		/* remember handle */
	XMM_Handle_Table[hnd].adj_rec_size = adj_recsize;	/* remember adj rec size */
	XMM_Handle_Table[hnd].act_rec_size = RecSize;		/* remember user rec size */
	XMM_Handle_Table[hnd].max_rec_alloc = MaxRecs;		/* remember size */
	XMM_Handle_Table[hnd].max_rec_used = -1L;				/* nothing yet */

	if (hnd > Max_Handle_Used)						/* exceed prev max */
		Max_Handle_Used = hnd;						/* yes */

/*************************************************************************
*                                                                        *
* 	Clear the memory																		 *
*                                                                        *
*************************************************************************/

	memset((char *)block, '\0', sizeof(block));

	for (i=0; i < space_needed; i++)
		{
		CtrlBlk.Length = sizeof(block);				/* bytes to move */
		CtrlBlk.SourceHandle = 0;						/* from real memory */
		(char far *)CtrlBlk.SourceOffset = (char far *)block;
		CtrlBlk.DestHandle = XMM_Handle_Table[hnd].handle;
		CtrlBlk.DestOffset = i * ONE_K;				/* dest */
		ret_code = XMM_MoveExtended(&CtrlBlk);		/* do it */
		if (ret_code < 0L)								/* error?? */
			{
			XMMErrorCode = XMSERROR(ret_code);		/* yes - save error */
			return (-4);									/* and leave */
			}
		}

/*************************************************************************
*                                                                        *
* 	All Done																					 *
*                                                                        *
*************************************************************************/

	return (hnd);

}

/*<f>*/
int DeAllocateXMMArea(XHandle hnd)
/*************************************************************************
*                                                                        *
* 	Delalocate an XMM area                                            	 *
*                                                                        *
*************************************************************************/
{
	long ret_code;

	XMMErrorCode = 0;									/* clear code */

	if (!initialized)									/* started?? */
		return (-1);									/* no */

	if (XMM_Handle_Table[hnd].handle == UNUSED)/* allocated?? */
		return (-2);									/* no */

	ret_code = XMM_FreeExtended(XMM_Handle_Table[hnd].handle);
	if (ret_code < 0L)								/* error?? */
		{
		XMMErrorCode = XMSERROR(ret_code);		/* yes - save error */
		return (-3);									/* and leave */
		}

	XMM_Handle_Table[hnd].handle = UNUSED;		/* been dropped */

	return (S_OKAY);

}

/*<f>*/
int GetXMMRecord(XHandle hnd, long recno, void *buffer)
/*************************************************************************
*                                                                        *
* 	Get a record from the XMM area pointer to by handle                	 *
*                                                                        *
*  Note - recno of 0 is valid                                            *
*                                                                        *
*************************************************************************/
{
	long ret_code;
	struct	XMM_Move CtrlBlk;

	XMMErrorCode = 0;									/* clear code */

	if (!initialized)									/* started?? */
		return (-1);									/* no */

	if (XMM_Handle_Table[hnd].handle == UNUSED)/* allocated?? */
		return (-2);									/* no */

	if (recno < 0L || recno > XMM_Handle_Table[hnd].max_rec_used)
		return (-4);

	CtrlBlk.Length = XMM_Handle_Table[hnd].adj_rec_size;
	CtrlBlk.SourceHandle = XMM_Handle_Table[hnd].handle;
	CtrlBlk.SourceOffset = recno * (XMM_Handle_Table[hnd].adj_rec_size);
	CtrlBlk.DestHandle = 0;
	(char far *)CtrlBlk.DestOffset = (char far *)move_area;

	ret_code = XMM_MoveExtended(&CtrlBlk);		/* get it */
	if (ret_code < 0L)								/* error?? */
		{
		XMMErrorCode = XMSERROR(ret_code);		/* yes - save error */
		return (-5);									/* and leave */
		}

	memmove(buffer, move_area, XMM_Handle_Table[hnd].act_rec_size);

	return (S_OKAY);

}

/*<f>*/
int CheckXMMRecord(XHandle hnd, long recno)
/*************************************************************************
*                                                                        *
* 	Check if a record exists - if so return S_OKAY else S_NOTFOUND      	 *
*                                                                        *
*  Note - recno of 0 is valid                                            *
*                                                                        *
*************************************************************************/
{
	long ret_code;
	struct	XMM_Move CtrlBlk;

	XMMErrorCode = 0;									/* clear code */

	if (!initialized)									/* started?? */
		return (-1);									/* no */

	if (XMM_Handle_Table[hnd].handle == UNUSED)/* allocated?? */
		return (-2);									/* no */

	if (recno < 0L)
		return (-4);

	if (recno > XMM_Handle_Table[hnd].max_rec_used)
		return (S_NOTFOUND);

	CtrlBlk.Length = XMM_Handle_Table[hnd].adj_rec_size;
	CtrlBlk.SourceHandle = XMM_Handle_Table[hnd].handle;
	CtrlBlk.SourceOffset = recno * XMM_Handle_Table[hnd].adj_rec_size;
	CtrlBlk.DestHandle = 0;
	(char far *)CtrlBlk.DestOffset = (char far *)move_area;

	ret_code = XMM_MoveExtended(&CtrlBlk);		/* get it */
	if (ret_code < 0L)								/* error?? */
		{
		XMMErrorCode = XMSERROR(ret_code);		/* yes - save error */
		return (-5);									/* and leave */
		}

	if (null_record(move_area, XMM_Handle_Table[hnd].act_rec_size))
		return (S_NOTFOUND);

	return (S_OKAY);

}

/*<f>*/
int PutXMMRecord(XHandle hnd, long recno, void *buffer)
/*************************************************************************
*                                                                        *
* 	Put a record into the XMM area pointer to by handle                	 *
*                                                                        *
*  Note - recno of 0 is valid                                            *
*                                                                        *
*************************************************************************/
{
	long ret_code;
	struct	XMM_Move CtrlBlk;

	XMMErrorCode = 0;									/* clear code */

	if (!initialized)									/* started?? */
		return (-1);									/* no */

	if (XMM_Handle_Table[hnd].handle == UNUSED)/* allocated?? */
		return (-2);									/* no */

	if (recno < 0L || recno > XMM_Handle_Table[hnd].max_rec_alloc)
		return (-4);

	memmove(move_area, buffer, XMM_Handle_Table[hnd].act_rec_size);

	CtrlBlk.Length = XMM_Handle_Table[hnd].adj_rec_size;
	CtrlBlk.SourceHandle = 0;
	(char far *)CtrlBlk.SourceOffset = (char far *)move_area;
	CtrlBlk.DestHandle = XMM_Handle_Table[hnd].handle;
	CtrlBlk.DestOffset = recno * XMM_Handle_Table[hnd].adj_rec_size;

	ret_code = XMM_MoveExtended(&CtrlBlk);		/* get it */
	if (ret_code < 0L)								/* error?? */
		{
		XMMErrorCode = XMSERROR(ret_code);		/* yes - save error */
		return (-5);									/* and leave */
		}

	if (recno > XMM_Handle_Table[hnd].max_rec_used)
		XMM_Handle_Table[hnd].max_rec_used = recno;

	return (S_OKAY);

}

/*<f>*/
static int null_record(char *buffer, int buff_len)
/*************************************************************************
*                                                                        *
* 	See if a record area is empty (NULL)    		                     	 *
*                                                                        *
*************************************************************************/
{
	int i;

	while (buff_len--)
		if (*(buffer++) != '\0')					/* anything there?? */
			return (FALSE);

	return (TRUE);

}
